#!/usr/bin/env python

# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import sys
import os.path
import glob
import argparse

import database.io as io
import database.db as db
import database.clblast as clblast
import database.bests as bests
import database.defaults as defaults

# Server storing a copy of the database
DATABASE_SERVER_URL = "https://raw.githubusercontent.com/CNugteren/CLBlast-database/master/database.json"


def remove_mismatched_arguments(database):
    """Checks for tuning results with mis-matched entries and removes them according to user preferences"""
    kernel_attributes = clblast.DEVICE_TYPE_ATTRIBUTES + clblast.KERNEL_ATTRIBUTES + ["kernel"]

    # For Python 2 and 3 compatibility
    try:
        user_input = raw_input
    except NameError:
        user_input = input
        pass

    # Check for mis-matched entries
    for kernel_group_name, kernel_group in db.group_by(database["sections"], kernel_attributes):
        group_by_arguments = db.group_by(kernel_group, clblast.ARGUMENT_ATTRIBUTES)
        if len(group_by_arguments) != 1:
            print("[database] WARNING: entries for a single kernel with multiple argument values " +
                  str(kernel_group_name))
            print("[database] Either quit or remove all but one of the argument combinations below:")
            for index, (attribute_group_name, mismatching_entries) in enumerate(group_by_arguments):
                print("[database]     %d: %s" % (index, attribute_group_name))
            for attribute_group_name, mismatching_entries in group_by_arguments:
                response = user_input("[database] Remove entries corresponding to %s, [y/n]? " %
                                      str(attribute_group_name))
                if response == "y":
                    for entry in mismatching_entries:
                        database["sections"].remove(entry)
                    print("[database] Removed %d entry/entries" % len(mismatching_entries))

    # Sanity-check: all mis-matched entries should be removed
    for kernel_group_name, kernel_group in db.group_by(database["sections"], kernel_attributes):
        group_by_arguments = db.group_by(kernel_group, clblast.ARGUMENT_ATTRIBUTES)
        if len(group_by_arguments) != 1:
            print("[database] ERROR: entries for a single kernel with multiple argument values " +
                  str(kernel_group_name))
        assert len(group_by_arguments) == 1


def remove_database_entries(database, remove_if_matches_fields):
    assert len(remove_if_matches_fields.keys()) > 0

    def remove_this_entry(section):
        for key in remove_if_matches_fields.keys():
            if section[key] != remove_if_matches_fields[key]:
                return False
        return True

    old_length = len(database["sections"])
    database["sections"] = [x for x in database["sections"] if not remove_this_entry(x)]
    new_length = len(database["sections"])
    print("[database] Removed %d entries from the database" % (old_length - new_length))


def add_tuning_parameter(database, parameter_name, kernel, value):
    num_changes = 0
    for section in database["sections"]:
        if section["kernel"] == kernel:
            for result in section["results"]:
                if parameter_name not in result["parameters"]:
                    result["parameters"][parameter_name] = value
            section["parameter_names"].append(parameter_name)
            num_changes += 1
    print("[database] Made %d addition(s) of %s" % (num_changes, parameter_name))


def main(argv):

    # Parses the command-line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("source_folder", help="The folder with JSON files to parse to add to the database")
    parser.add_argument("clblast_root", help="Root of the CLBlast sources")
    parser.add_argument("-r", "--remove_device", type=str, default=None, help="Removes all entries for a specific device")
    parser.add_argument("--add_tuning_parameter", type=str, default=None, help="Adds this parameter to existing entries")
    parser.add_argument("--add_tuning_parameter_for_kernel", type=str, default=None, help="Adds the above parameter for this kernel")
    parser.add_argument("--add_tuning_parameter_value", type=int, default=0, help="Set this value as the default for the above parameter")
    parser.add_argument("-v", "--verbose", action="store_true", help="Increase verbosity of the script")
    cl_args = parser.parse_args(argv)

    # Parses the path arguments
    database_filename = os.path.join(cl_args.clblast_root, "scripts", "database", "database.json")
    database_best_filename = os.path.join(cl_args.clblast_root, "scripts", "database", "database_best.json")
    json_files = os.path.join(cl_args.source_folder, "*.json")
    cpp_database_path = os.path.join(cl_args.clblast_root, "src", "database", "kernels")

    # Checks whether the command-line arguments are valid
    clblast_header = os.path.join(cl_args.clblast_root, "include", "clblast.h")  # Not used but just for validation
    if not os.path.isfile(clblast_header):
        raise RuntimeError("The path '" + cl_args.clblast_root +
                           "' does not point to the root of the CLBlast library")
    if len(glob.glob(json_files)) < 1:
        print("[database] The path '" + cl_args.source_folder + "' does not contain any JSON files")

    # Downloads the database if a local copy is not present
    if not os.path.isfile(database_filename):
        io.download_database(database_filename, DATABASE_SERVER_URL)

    # Loads the database from disk
    database = io.load_database(database_filename)

    # Loops over all JSON files in the supplied folder
    for file_json in glob.glob(json_files):
        sys.stdout.write("[database] Processing '" + file_json + "' ")  # No newline printed

        try:
            # Loads the newly imported data
            imported_data = io.load_tuning_results(file_json)

            # Adds the new data to the database
            old_size = db.length(database)
            database = db.add_section(database, imported_data)
            new_size = db.length(database)
            print("with " + str(new_size - old_size) + " new items")  # Newline printed here

        except ValueError:
            print("--- WARNING: invalid file, skipping")

    # Checks for tuning results with mis-matched entries
    remove_mismatched_arguments(database)

    # Stores the modified database back to disk
    if len(glob.glob(json_files)) >= 1:
        io.save_database(database, database_filename)

    # Removes database entries before continuing
    if cl_args.remove_device is not None:
        print("[database] Removing all results for device '%s'" % cl_args.remove_device)
        remove_database_entries(database, {"clblast_device_name": cl_args.remove_device})
                                           #, "kernel_family": "xgemm"})
        io.save_database(database, database_filename)

    # Adds new tuning parameters to existing database entries
    if cl_args.add_tuning_parameter is not None and\
       cl_args.add_tuning_parameter_for_kernel is not None:
        print("[database] Adding tuning parameter: '%s' for kernel '%s' with default %d" %
              (cl_args.add_tuning_parameter, cl_args.add_tuning_parameter_for_kernel,
               cl_args.add_tuning_parameter_value))
        add_tuning_parameter(database, cl_args.add_tuning_parameter,
                             cl_args.add_tuning_parameter_for_kernel,
                             cl_args.add_tuning_parameter_value)
        io.save_database(database, database_filename)

    # Retrieves the best performing results
    print("[database] Calculating the best results per device/kernel...")
    database_best_results = bests.get_best_results(database)

    # Determines the defaults for other vendors and per vendor
    print("[database] Calculating the default values...")
    database_defaults = defaults.calculate_defaults(database, cl_args.verbose)
    database_best_results["sections"].extend(database_defaults["sections"])

    # Optionally outputs the database to disk
    if cl_args.verbose:
        io.save_database(database_best_results, database_best_filename)

    # Outputs the database as a C++ database
    print("[database] Producing a C++ database in '" + cpp_database_path + "'...")
    clblast.print_cpp_database(database_best_results, cpp_database_path)

    print("[database] All done")


if __name__ == '__main__':
    main(sys.argv[1:])



# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import utils

import matplotlib
matplotlib.use('Agg')
from matplotlib import rcParams
import matplotlib.pyplot as plt
import numpy as np

# Colors
BLUEISH = [c / 255.0 for c in [71, 101, 177]]  # #4765b1
REDISH = [c / 255.0 for c in [214, 117, 104]]  # #d67568
PURPLISH = [c / 255.0 for c in [85, 0, 119]]  # #550077
GREEN = [c / 255.0 for c in [144, 224, 98]] # #90e062
COLORS = [BLUEISH, REDISH, PURPLISH, GREEN]
MARKERS = ["o-", "x-", ".-",  "--"]


def plot_graphs(results, file_name, num_rows, num_cols,
                x_keys, y_keys, titles, x_labels, y_labels,
                label_names, title, tight_plot, verbose):
    assert len(results) == num_rows * num_cols
    assert len(results) >= 1
    assert len(x_keys) == len(results)
    assert len(y_keys) == len(results)
    assert len(titles) == len(results)
    assert len(x_labels) == len(results)
    assert len(y_labels) == len(results)

    # Tight plot (for in a paper or presentation) or regular (for display on a screen)
    if tight_plot:
        plot_size = 5
        w_space = 0.20
        h_space = 0.39
        title_from_top = 0.11
        legend_from_top = 0.17
        legend_from_top_per_item = 0.04
        x_label_from_bottom = 0.09
        legend_spacing = 0.0
        font_size = 15
        font_size_legend = 13
        font_size_title = font_size
        bounding_box = "tight"
    else:
        plot_size = 8
        w_space = 0.15
        h_space = 0.22
        title_from_top = 0.09
        legend_from_top = 0.10
        legend_from_top_per_item = 0.07
        x_label_from_bottom = 0.06
        legend_spacing = 0.8
        font_size = 15
        font_size_legend = font_size
        font_size_title = 18
        bounding_box = None  # means not 'tight'

    # Initializes the plot
    size_x = plot_size * num_cols
    size_y = plot_size * num_rows
    rcParams.update({'font.size': font_size})
    fig, axes = plt.subplots(nrows=num_rows, ncols=num_cols, figsize=(size_x, size_y), facecolor='w', edgecolor='k')
    if len(results) == 1 and not type(axes) is np.ndarray:
        axes = np.full((1,1), axes)
    assert type(axes) is np.ndarray
    fig.text(.5, 0.92, title, horizontalalignment="center", fontsize=font_size_title)
    plt.subplots_adjust(wspace=w_space, hspace=h_space)

    # Loops over each subplot
    for row in range(num_rows):
        for col in range(num_cols):
            index = row * num_cols + col
            result = results[index]
            if num_rows == 1:
                ax = axes[col]
            elif num_cols == 1:
                ax = axes[row]
            else:
                ax = axes[row, col]
            plt.sca(ax)
            print("[plot] Plotting subplot %d" % index)

            # Sets the x-axis labels
            x_list = [[r[x_key] for r in result] for x_key in x_keys[index]]
            x_ticks = [",".join([utils.float_to_kilo_mega(v) for v in values]) for values in zip(*x_list)]
            x_location = range(len(x_ticks))

            # Optional sparsifying of the labels on the x-axis
            if tight_plot and len(x_location) > 10:
                x_ticks = [v if not (i % 2) else "" for i, v in enumerate(x_ticks)]

            # Sets the y-data
            y_list = [[r[y_key] if y_key in r.keys() and not isinstance(r[y_key], str) else 0 for r in result]
                      for y_key in y_keys[index]]
            y_max = [max(y) if len(y) else 1 for y in y_list]
            y_max = max(y_max) if len(y_list) > 0 else 1

            # Sets the axes
            y_rounding = 10 if y_max < 80 else 50 if y_max < 400 else 200
            y_axis_limit = (y_max * 1.2) - ((y_max * 1.2) % y_rounding) + y_rounding
            plt.ylim(ymin=0, ymax=y_axis_limit)
            plt.xticks(x_location, x_ticks, rotation='vertical')

            # Sets the labels
            ax.set_title(titles[index], y=1.0 - title_from_top, fontsize=font_size)
            if col == 0 or y_labels[index] != y_labels[index - 1]:
                ax.set_ylabel(y_labels[index])
            ax.set_xlabel(x_labels[index])
            ax.xaxis.set_label_coords(0.5, x_label_from_bottom)

            # Plots the graph
            assert len(COLORS) >= len(y_keys[index])
            assert len(MARKERS) >= len(y_keys[index])
            assert len(label_names) == len(y_keys[index])
            for i in range(len(y_keys[index])):
                color = COLORS[i]
                marker = MARKERS[i]
                if label_names[i] in ["CLBlast", "CLBlast FP32"]:
                    color = BLUEISH
                    marker = "o-"
                elif label_names[i] in ["CLBlast FP16"]:
                    color = PURPLISH
                    marker = ".-"
                elif label_names[i] in ["clBLAS", "clBLAS FP32", "clBLAS (non-batched)"]:
                    color = REDISH
                    marker = "x-"
                elif label_names[i] in ["cuBLAS", "cuBLAS (non-batched)"]:
                    color = GREEN
                    marker = ".-"
                ax.plot(x_location, y_list[i], marker, label=label_names[i], color=color)

            # Sets the legend
            leg = ax.legend(loc=(0.02, 1.0 - legend_from_top - legend_from_top_per_item * len(y_keys[index])),
                            handletextpad=0.1, labelspacing=legend_spacing, fontsize=font_size_legend)
            leg.draw_frame(False)

    # Saves the plot to disk
    print("[benchmark] Saving plot to '" + file_name + "'")
    fig.savefig(file_name, bbox_inches=bounding_box)




####################################################################################################
# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>
#
# This file test PyCLBlast: the Python interface to CLBlast. It is not exhaustive. For full testing
# it is recommended to run the regular CLBlast tests, this is just a small smoke test.
#
####################################################################################################

import unittest

import numpy as np
import pyopencl as cl
from pyopencl.array import Array

import pyclblast


class TestPyCLBlast(unittest.TestCase):

    @staticmethod
    def setup(sizes, dtype):
        ctx = cl.create_some_context()
        queue = cl.CommandQueue(ctx)
        host_arrays, device_arrays = [], []
        for size in sizes:
            numpy_array = np.random.rand(*size).astype(dtype=dtype)
            opencl_array = Array(queue, numpy_array.shape, numpy_array.dtype)
            opencl_array.set(numpy_array)
            host_arrays.append(numpy_array)
            device_arrays.append(opencl_array)
        queue.finish()
        return queue, host_arrays, device_arrays

    def test_axpy(self):
        for dtype in ["float32", "complex64"]:
            for alpha in [1.0, 3.1]:
                for n in [1, 7, 32]:
                    queue, h, d = self.setup([(n,), (n,)], dtype=dtype)
                    pyclblast.axpy(queue, n, d[0], d[1], alpha=alpha)
                    queue.finish()
                    result = d[1].get()
                    reference = alpha * h[0] + h[1]
                    for i in range(n):
                        self.assertAlmostEqual(reference[i], result[i], places=3)

    def test_gemv(self):
        for dtype in ["float32", "complex64"]:
            for beta in [1.0]:
                for alpha in [1.0, 3.1]:
                    for m in [1, 7, 32]:
                        for n in [1, 7, 32]:
                            queue, h, d = self.setup([(m, n), (n,), (m,)], dtype=dtype)
                            pyclblast.gemv(queue, m, n, d[0], d[1], d[2],
                                           a_ld=n, alpha=alpha, beta=beta)
                            queue.finish()
                            result = d[2].get()
                            reference = alpha * np.dot(h[0], h[1]) + beta * h[2]
                            for i in range(m):
                                self.assertAlmostEqual(reference[i], result[i], places=3)

    def test_gemm(self):
        for dtype in ["float32", "complex64"]:
            for beta in [1.0]:
                for alpha in [1.0, 3.1]:
                    for m in [1, 7, 32]:
                        for n in [1, 7, 32]:
                            for k in [1, 7, 32]:
                                queue, h, d = self.setup([(m, k), (k, n), (m, n)], dtype=dtype)
                                pyclblast.gemm(queue, m, n, k, d[0], d[1], d[2],
                                               a_ld=k, b_ld=n, c_ld=n, alpha=alpha, beta=beta)
                                queue.finish()
                                result = d[2].get()
                                reference = alpha * np.dot(h[0], h[1]) + beta * h[2]
                                for i in range(m):
                                    for j in range(n):
                                        self.assertAlmostEqual(reference[i, j], result[i, j],
                                                               places=3)




# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import re
import json

try:
    from urllib.request import urlopen  # Python 3
except ImportError:
    from urllib2 import urlopen  # Python 2


def download_database(filename, database_url):
    """Downloads a database and saves it to disk"""
    print("[database] Downloading database from '" + database_url + "'...")
    database = urlopen(database_url)
    with open(filename, "wb") as f:
        f.write(database.read())


def load_database(filename):
    """Loads a database from disk"""
    print("[database] Loading database from '" + filename + "'")
    with open(filename) as f:
        database = json.load(f)
    return decompress_database(database)


def save_database(database, filename):
    """Saves a database to disk"""
    compressed_db = compress_database(database)
    print("[database] Saving database to '" + filename + "'")
    with open(filename, "w") as f:
        json.dump(compressed_db, f, sort_keys=True, indent=2, separators=(',', ': '))


def compress_database(database):
    """Moves certain common fields up in the hierarchy, transforms dicts into lists"""
    new_sections = []
    for section in database["sections"]:
        new_section = {}
        for field in section:
            if field == "results":
                parameter_names = [sorted(result["parameters"].keys()) for result in section["results"]]
                assert len(list(set([" ".join(p) for p in parameter_names]))) == 1
                new_section["parameter_names"] = parameter_names[0]  # they are all the same
                new_results = [[",".join([str(result["parameters"][p]) for p in new_section["parameter_names"]]),
                                result["time"]]
                               for result in section["results"]]
                new_section[field] = new_results
            elif field != "parameter_names":
                new_section[field] = section[field]
        new_sections.append(new_section)
    return {"sections": new_sections}


def decompress_database(database):
    """Undo the above compression"""
    for section in database["sections"]:
        new_results = []
        for result in section["results"]:
            parameters = {}
            for name, value in zip(section["parameter_names"], result[0].split(",")):
                parameters[name] = int(value)
            new_result = {
                "parameters": parameters,
                "time": result[1]
            }
            new_results.append(new_result)
        section["results"] = new_results
    return database


def load_tuning_results(filename):
    """Loads JSON data from file and pre-processes it"""
    with open(filename) as f:
        json_data = json.load(f)

    # Removes the numbering following the kernel family name
    json_data["kernel_family"] = re.sub(r'_\d+', '', json_data["kernel_family"])

    # Removes unnecessary data
    if json_data["best_kernel"]:
        del json_data["best_kernel"]
    if json_data["best_time"]:
        del json_data["best_time"]
    if json_data["best_parameters"]:
        del json_data["best_parameters"]

    # Adds the kernel name to the section instead of to the individual results
    assert len(json_data["results"]) > 0
    json_data["kernel"] = json_data["results"][0]["kernel"]
    for result in json_data["results"]:
        assert json_data["kernel"] == result["kernel"]
        result.pop("kernel", None)

    # Removes the 'PRECISION' parameter from the individual results: it is redundant
    for result in json_data["results"]:
        assert json_data["precision"] == str(result["parameters"]["PRECISION"])
        result["parameters"].pop("PRECISION", None)

    # Fixes the scalar argument values
    for value, replacement in zip(["2.00", "2.00+0.50i"], ["2.000000", "2+0.5i"]):
        for field in ["arg_alpha", "arg_beta"]:
            if field in json_data.keys() and json_data[field] == value:
                json_data[field] = replacement

    # All done
    return json_data



#!/usr/bin/env python

# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0.
# This file follows the PEP8 Python style guide and uses a max-width of 100 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import numpy as np
import pyopencl as cl
from pyopencl.array import Array
import pyclblast

# Settings for this sample
dtype = 'float32'

print("# Setting up OpenCL")
ctx = cl.create_some_context()
queue = cl.CommandQueue(ctx)

print("# Setting up Numpy arrays")
m, n, k = 2, 3, 4
a = np.random.rand(m, k).astype(dtype=dtype)
b = np.random.rand(k, n).astype(dtype=dtype)
c = np.random.rand(m, n).astype(dtype=dtype)

print("# Setting up OpenCL arrays")
cla = Array(queue, a.shape, a.dtype)
clb = Array(queue, b.shape, b.dtype)
clc = Array(queue, c.shape, c.dtype)
cla.set(a)
clb.set(b)
clc.set(c)

print("# Example level-3 operation: GEMM")
pyclblast.gemm(queue, m, n, k, cla, clb, clc, a_ld=k, b_ld=n, c_ld=n)
queue.finish()
print("# Matrix C result: %s" % clc.get())
print("# Expected result: %s" % (np.dot(a, b)))




# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import ast
from collections import defaultdict

import database.bests as bests
import database.clblast as clblast


def set_identifiers(database, group_by_attributes, identifier_name):
    """Sets a group-identifier based on a given set of attributes. Modifies the database but also returns a list of
    unique identifiers."""
    identifiers = []
    for section in database["sections"]:
        identifier = []
        for attribute in group_by_attributes:
            if attribute in section:
                identifier.append(section[attribute])
        section[identifier_name] = ";".join(identifier)
        identifiers.append(section[identifier_name])
    return sorted(set(identifiers))


def remove_identifiers(database, identifier_name):
    """Removes an identifier from all sections in the database"""
    for section in database["sections"]:
        section.pop(identifier_name, None)


def get_groups_by_identifier(database, group_identifiers, identifier_name):
    """Returns a list of (group, group_identifier) tuples based a previously made grouping"""
    groups = []
    for group_identifier in group_identifiers:

        # Get all sections in this group
        group = []
        for section in database["sections"]:
            if section[identifier_name] == group_identifier:
                group.append(section)

        groups.append((group, group_identifier))
    return groups


def add_default_sections(database, grouping, verbose, values_dict, condition, enable_warning):
    default_sections = []

    # Groups the database by a certain grouping
    group_identifiers = set_identifiers(database, grouping, "group_identifier")
    groups = get_groups_by_identifier(database, group_identifiers, "group_identifier")

    # Loops over all groups
    for group, group_identifier in groups:

        # Computes the best parameters
        default_parameters = get_common_best_parameters(group, group_identifier, verbose, enable_warning)
        assert len(group) > 0
        if condition(group[0]):

            # Stores all the section's data
            default_section = {}
            for attribute in group[0].keys():
                if attribute != "results" and attribute != "group_identifier":
                    default_section[attribute] = group[0][attribute]
            default_section["clblast_device_compute_units"] = 0
            default_section["clblast_device_core_clock"] = 0
            for key in values_dict.keys():
                default_section[key] = values_dict[key]
            default_section["results"] = [{"time": 0.0, "parameters": default_parameters}]
            default_sections.append(default_section)
    return default_sections


def calculate_defaults(database, verbose):
    """Sets defaults for devices of the same type/vendor"""
    default_sections = {"sections": []}

    # Groups the database by kernel, vendor and device architecture (e.g. AMD GPU "Fiji")
    architecture_group = clblast.GROUP_ATTRIBUTES + ["clblast_device_architecture"]
    architecture_defaults = add_default_sections(database, architecture_group, verbose,
                                                 {"clblast_device_name": clblast.DEVICE_NAME_DEFAULT},
                                                 lambda entry: True, enable_warning=False)

    # Groups the database by kernel, vendor and device type (e.g. AMD GPU)
    device_defaults = add_default_sections(database, clblast.GROUP_ATTRIBUTES, verbose,
                                           {"clblast_device_name": clblast.DEVICE_NAME_DEFAULT,
                                            "clblast_device_architecture": clblast.DEVICE_ARCHITECTURE_DEFAULT},
                                           lambda entry: entry["clblast_device_architecture"] != "",
                                           enable_warning=True)
    default_sections["sections"].extend(device_defaults)

    # Groups the database by kernel, vendor and device type (e.g. AMD GPU) - but not by arguments!
    # This is to check for mis-matched arguments in the database. Note: this is not a check on the
    # architecture defaults
    attributes = clblast.DEVICE_TYPE_ATTRIBUTES + clblast.KERNEL_ATTRIBUTES + ["kernel"]
    group_identifiers = set_identifiers(default_sections, attributes, "temp_identifier")
    groups = get_groups_by_identifier(default_sections, group_identifiers, "temp_identifier")
    for group, group_identifier in groups:
        if len(group) != 1:
            print("[ERROR] Entries for a single kernel with multiple argument values: " + str(group_identifier))
        assert len(group) == 1
    remove_identifiers(default_sections, "temp_identifier")

    # Adds the architecture defaults only after running the above check
    default_sections["sections"].extend(architecture_defaults)

    # Groups the database by kernel only
    group_identifiers = set_identifiers(database, clblast.KERNEL_ATTRIBUTES + ["kernel"], "group_identifier")
    groups = get_groups_by_identifier(database, group_identifiers, "group_identifier")

    # Loops over all groups
    for group, group_identifier in groups:

        # Computes the best parameters
        default_parameters = get_common_best_parameters(group, group_identifier, verbose,
                                                        enable_warning=True)

        # Stores all the section's data
        assert len(group) > 0
        default_section = {}
        for attribute in group[0].keys():
            if attribute != "results" and attribute != "group_identifier":
                default_section[attribute] = group[0][attribute]
        default_section["clblast_device_name"] = clblast.DEVICE_NAME_DEFAULT
        default_section["clblast_device_architecture"] = clblast.DEVICE_ARCHITECTURE_DEFAULT
        default_section["clblast_device_vendor"] = clblast.VENDOR_DEFAULT
        default_section["clblast_device_type"] = clblast.DEVICE_TYPE_DEFAULT
        default_section["clblast_device_compute_units"] = 0
        default_section["clblast_device_core_clock"] = 0
        default_section["results"] = [{"time": 0.0, "parameters": default_parameters}]
        default_sections["sections"].append(default_section)

    # Database with both types of defaults only
    return default_sections


def get_smallest_best_parameters(group):
    """Sets defaults based on the smallest values of all known entries. The average might be better for performance but
    some parameters might not be supported on other devices."""

    # Counts the number of devices in this group
    assert len(group) > 0

    # Find the smallest values of the parameters
    min_parameters = {}
    for section in group:
        assert len(section["results"]) > 0
        minimum_time = min([result["time"] for result in section["results"]])
        for result in section["results"]:
            if result["time"] == minimum_time:
                for parameter in result["parameters"]:
                    if parameter in min_parameters:
                        min_parameters[parameter] = min(min_parameters[parameter], result["parameters"][parameter])
                    else:
                        min_parameters[parameter] = result["parameters"][parameter]

    return min_parameters


def get_parameter_names(section):
    return [result["parameters"] for result in section["results"]]


def get_common_best_parameters(group, group_identifier, verbose, enable_warning):
    """Sets defaults based on the best values of entries supported by all devices. This might cause a problem in case
    not every device was tuned with the same parameters. In that case it falls back to the above method to retrieve
    the smallest best execution time"""

    # Counts the number of devices in this group
    num_devices = len(group)
    assert num_devices > 0

    # Inserts the relative execution times into the database
    for section in group:
        assert len(section["results"]) > 0
        minimum_time = min([result["time"] for result in section["results"]])
        for result in section["results"]:
            base_line = minimum_time if section["kernel"] != "gemm_kernel_selection" else 1.0
            result["relative_time"] = result["time"] / base_line

    # Determine which parameters are available for all devices
    common_parameters = get_parameter_names(group[0])  # Parameters of the first section
    for i in range(1, num_devices):
        section_parameters = get_parameter_names(group[i])
        common_parameters = [p for p in section_parameters if p in common_parameters]  # Intersection of the parameters

    # Fall back to another method in case there are no shared entries at all across devices
    if len(common_parameters) == 0:
        if verbose:
            print("[database] No common kernels for: " + str(group_identifier) + " across all %d devices " % num_devices)

        # Computes the amount of devices with shared parameters
        parameters_count = defaultdict(int)
        for i in range(0, num_devices):
            for parameters in get_parameter_names(group[i]):
                parameters_count[str(parameters)] += 1
        num_devices_common = max(parameters_count.values())

        # Fall back method in case there are no shared entries at all across devices
        if num_devices_common == 1:
            if enable_warning:
                print("[database] Warning: No common kernels for: " + str(group_identifier) + " at all")
            smallest_best_parameters = get_smallest_best_parameters(group)
            if verbose:
                print("[database] " + str(group_identifier))
            return smallest_best_parameters

        # Checks if perhaps there are many more shared parameters with a bit fewer devices
        num_parameters_common = defaultdict(int)
        for count in parameters_count.values():
            if count != 1:
                num_parameters_common[str(count)] += 1
        if num_parameters_common[str(num_devices_common - 1)] > num_parameters_common[str(num_devices_common)]:
            num_devices_common -= 1
        if verbose:
            print("[database] Found %d common kernels for: " % num_parameters_common[str(num_devices_common)] +
                  str(group_identifier) + " across %d out of %d devices " % (num_devices_common, num_devices))

        # Populates the common parameters
        for parameters_string in parameters_count.keys():
            count = parameters_count[parameters_string]
            if count == num_devices_common:
                parameters = ast.literal_eval(parameters_string)
                common_parameters.append(parameters)

    # Removes entries with parameters which are not common
    common_results = []
    for section in group:
        for result in section["results"]:
            if result["parameters"] in common_parameters:
                common_results.append(result)

    # Retrieves the entries with the highest relative performance
    relative_best_parameters = bests.get_relative_bests(group_identifier, common_results, common_parameters, verbose)
    return relative_best_parameters




# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import itertools
from operator import itemgetter


def length(database):
    """Computes the total number of tuning entries"""
    num_tuning_entries = 0
    for section in database["sections"]:
        num_tuning_entries += len(section["results"])
    return num_tuning_entries


def add_section(database, new_section):
    """Adds a new section to the database"""
    for old_section in database["sections"]:

        # Verify whether the sections match
        equal = True
        for attribute in new_section.keys():
            if attribute != "results":
                if attribute not in old_section or new_section[attribute] != old_section[attribute]:
                    equal = False
                    break

        # They match: append the new section's results to the corresponding entry in the database and return
        if equal:
            old_section["results"] = combine_results(old_section["results"], new_section["results"])
            return database

    # No match found: append the whole new section to the database
    database["sections"].append(new_section)
    return database


def combine_results(old_results, new_results):
    """Adds new results to the results JSON list"""
    for new_result in new_results:
        old_results = combine_result(old_results, new_result)
    return old_results


def combine_result(old_results, new_result):
    """Adds a new result to the results JSON list; filters for duplicate entries and saves the best performing one"""

    # Loops over all existing results to test for already existing entries with these parameters
    for old_result in old_results:

        # Verify whether the results match
        equal = new_result["parameters"] == old_result["parameters"]

        # They match: keep only the one with the minimum execution time
        if equal:
            old_result["time"] = min(old_result["time"], new_result["time"])
            return old_results

    # No match found: append a new result
    old_results.append(new_result)
    return old_results


def group_by(database, attributes):
    """Returns an list with the name of the group and the corresponding entries in the database"""
    assert len(database) > 0
    attributes = [a for a in attributes if a in database[0]]
    database.sort(key=itemgetter(*attributes))
    result = []
    for key, data in itertools.groupby(database, key=itemgetter(*attributes)):
        result.append((key, list(data)))
    return result



#!/usr/bin/env python

# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import argparse
import json
import os
import sys

import settings
import plot
import utils

EXPERIMENTS = {
    "axpy": settings.AXPY,
    "axpybatched": settings.AXPYBATCHED,
    "gemv": settings.GEMV,
    "gemm": settings.GEMM,
    "gemm_small": settings.GEMM_SMALL,
    "gemmbatched": settings.GEMMBATCHED,
    "gemmstridedbatched": settings.GEMMSTRIDEDBATCHED,
    "symm": settings.SYMM,
    "syrk": settings.SYRK,
    "summary": settings.SUMMARY,
}

COMPARISONS = ["clBLAS", "CPU-BLAS", "cuBLAS"]
COMPARISON_ARGS = ["-clblas", "-cblas", "-cublas"]
COMPARISON_IDS = [2, 3, 4]


def run_benchmark(name, arguments_list, precision, num_runs, platform, device, comparisons):
    binary = "./clblast_client_x" + name

    # Loops over sub-benchmarks per benchmark
    results = []
    for arguments in arguments_list:

        # Sets the arguments
        constant_arguments = ["-warm_up", "-q", "-no_abbrv"]
        common_arguments = ["-precision %d" % precision, "-runs %d" % num_runs]
        opencl_arguments = ["-platform %d" % platform, "-device %d" % device]
        comparison_arguments = []
        for name, arg in zip(COMPARISONS, COMPARISON_ARGS):
            if name in comparisons:
                comparison_arguments.append(arg + " 1")
            else:
                comparison_arguments.append(arg + " 0")
        all_arguments = opencl_arguments + common_arguments + constant_arguments + comparison_arguments
        for name, value in arguments.items():
            all_arguments.append("-" + name + " " + str(value))

        # Calls the binary and parses the results
        benchmark_output = utils.run_binary(binary, all_arguments)
        result = utils.parse_results(benchmark_output)

        # For half-precision: also runs single-precision for comparison
        if precision == 16:
            all_arguments = [arg if arg != "-precision 16" else "-precision 32" for arg in all_arguments]
            benchmark_output = utils.run_binary(binary, all_arguments)
            result_extra = utils.parse_results(benchmark_output)
            for index in range(min(len(result), len(result_extra))):
                result[index]["GBs_1_FP32"] = result_extra[index]["GBs_1"]
                result[index]["GFLOPS_1_FP32"] = result_extra[index]["GFLOPS_1"]
                for id in COMPARISON_IDS:
                    if "GBs_%d" % id in result_extra[index].keys():
                        result[index]["GBs_%d" % id] = result_extra[index]["GBs_%d" % id]
                        result[index]["GFLOPS_%d" % id] = result_extra[index]["GFLOPS_%d" % id]

        results.extend(result)
    return results


def parse_arguments(argv):
    parser = argparse.ArgumentParser(description="Runs a full benchmark for a specific routine on a specific device")
    parser.add_argument("-b", "--benchmark", required=True, help="The benchmark to perform (choose from %s)" % sorted(EXPERIMENTS.keys()))
    parser.add_argument("-c", "--comparisons", default=[], nargs='+', help="The library(s) to compare against (choose from %s)" % COMPARISONS)
    parser.add_argument("-p", "--platform", required=True, type=int, help="The ID of the OpenCL platform to test on")
    parser.add_argument("-d", "--device", required=True, type=int, help="The ID of the OpenCL device to test on")
    parser.add_argument("-n", "--num_runs", type=int, default=None, help="Overrides the default number of benchmark repeats for averaging")
    parser.add_argument("-x", "--precision", type=int, default=32, help="The precision to test for (choose from 16, 32, 64, 3232, 6464")
    parser.add_argument("-l", "--load_from_disk", action="store_true", help="Loading existing results from JSON file and replot")
    parser.add_argument("-t", "--plot_title", default="", help="The title for the plots, defaults to benchmark name")
    parser.add_argument("-z", "--tight_plot", action="store_true", help="Enables tight plot layout for in paper or presentation")
    parser.add_argument("-o", "--output_folder", default=os.getcwd(), help="Sets the folder for output plots (defaults to current folder)")
    parser.add_argument("-v", "--verbose", action="store_true", help="Increase verbosity of the script")
    cl_args = parser.parse_args(argv)
    return vars(cl_args)


def benchmark_single(benchmark, comparisons, platform, device, num_runs, precision, load_from_disk,
                     plot_title, tight_plot, output_folder, verbose):

    # Sanity check
    if not os.path.isdir(output_folder):
        print("[benchmark] Error: folder '%s' doesn't exist" % output_folder)
        return

    # The benchmark name and plot title
    benchmark_name = utils.precision_to_letter(precision) + benchmark.upper()
    if benchmark.upper() != "SUMMARY":
        plot_title = benchmark_name if plot_title == "" else benchmark_name + ": " + plot_title

    # Retrieves the comparison settings
    library_ids = [1]
    for comparison in comparisons:
        if comparison not in COMPARISONS:
            print("[benchmark] Invalid comparison library '%s', choose from %s" % (comparison, COMPARISONS))
            return
        library_ids.append(COMPARISON_IDS[COMPARISONS.index(comparison)])

    # Retrieves the benchmark settings
    if benchmark not in EXPERIMENTS.keys():
        print("[benchmark] Invalid benchmark '%s', choose from %s" % (benchmark, EXPERIMENTS.keys()))
        return
    experiment = EXPERIMENTS[benchmark]
    benchmarks = experiment["benchmarks"]

    # Either run the benchmarks for this experiment or load old results from disk
    json_file_name = os.path.join(output_folder, benchmark_name.lower() + "_benchmarks.json")
    if load_from_disk and os.path.isfile(json_file_name):
        print("[benchmark] Loading previous benchmark results from '" + json_file_name + "'")
        with open(json_file_name) as f:
            results = json.load(f)
    else:

        # Runs all the individual benchmarks
        print("[benchmark] Running on platform %d, device %d" % (platform, device))
        print("[benchmark] Running %d benchmarks for settings '%s'" % (len(benchmarks), benchmark))
        results = {"label_names": ["CLBlast"] + comparisons, "num_rows": experiment["num_rows"],
                   "num_cols": experiment["num_cols"], "benchmarks": []}
        for bench in benchmarks:
            num_runs_benchmark = bench["num_runs"] if num_runs is None else num_runs
            print("[benchmark] Running benchmark '%s:%s'" % (bench["name"], bench["title"]))
            result = run_benchmark(bench["name"], bench["arguments"], precision, num_runs_benchmark,
                                   platform, device, comparisons)
            results["benchmarks"].append(result)

        # Stores the results to disk
        print("[benchmark] Saving benchmark results to '" + json_file_name + "'")
        with open(json_file_name, "w") as f:
            json.dump(results, f, sort_keys=True, indent=4)

    # Retrieves the data from the benchmark settings
    file_name_suffix = "_tight" if tight_plot else ""
    pdf_file_name = os.path.join(output_folder, benchmark_name.lower() + "_plot" + file_name_suffix + ".pdf")
    titles = [b["title"] if "BATCHED" in b["name"].upper() else
              utils.precision_to_letter(precision) + b["name"].upper() + " " + b["title"]
              for b in benchmarks]
    x_keys = [b["x_keys"] for b in benchmarks]
    y_keys = [["%s_%d" % (b["y_key"], i) for i in library_ids] for b in benchmarks]
    x_labels = [b["x_label"] for b in benchmarks]
    y_labels = [b["y_label"] for b in benchmarks]
    label_names = results["label_names"]

    # For half-precision: also adds single-precision results for comparison
    if precision == 16:
        label_names[0] += " FP16"
        for index in range(1, len(label_names)):
            label_names[index] += " FP32"
        label_names.append("CLBlast FP32")
        y_keys = [y_key + [y_key[0] + "_FP32"] for y_key in y_keys]

    # For batched routines: comparison is non-batched
    if benchmark in ["axpybatched", "gemmbatched", "gemmstridedbatched"]:
        for index in range(1, len(label_names)):
            label_names[index] += " (non-batched)"

    # Plots the graphs
    plot.plot_graphs(results["benchmarks"], pdf_file_name, results["num_rows"], results["num_cols"],
                     x_keys, y_keys, titles, x_labels, y_labels,
                     label_names, plot_title, tight_plot, verbose)

    print("[benchmark] All done")


if __name__ == '__main__':
    parsed_arguments = parse_arguments(sys.argv[1:])
    benchmark_single(**parsed_arguments)



#!/usr/bin/env python

# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0.
# This file follows the PEP8 Python style guide and uses a max-width of 100 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import numpy as np
import pyopencl as cl
from pyopencl.array import Array
import pyclblast
from datetime import datetime

if __name__ == "__main__":

    # Set up pyopencl:
    ctx = cl.create_some_context()
    queue = cl.CommandQueue(ctx)

    # Set up a basic sgemm example:
    m, n, k = 2, 3, 4
    a = np.random.rand(m, k).astype(dtype=np.float32)
    b = np.random.rand(k, n).astype(dtype=np.float32)
    c = np.empty((m, n), np.float32)
    cla = Array(queue, a.shape, a.dtype)
    clb = Array(queue, b.shape, b.dtype)
    clc = Array(queue, c.shape, c.dtype)
    cla.set(a)
    clb.set(b)
    clc.set(c)

    # Perform sgemm on these matrices, overriding the CLBlast parameters. In this example, we'll
    # just change the 'MWG' parameter a couple of times:
    params = { "KWG": 32, "KWI": 2, "MDIMA": 8, "MDIMC": 8, "MWG": 64, "NDIMB": 8, "NDIMC": 8,
            "NWG": 64, "SA": 0, "SB": 0, "STRM": 0, "STRN": 0, "VWM": 4, "VWN": 1 }
    for mwg in (32, 64, 256):
        print("Running sgemm tuned with MWG = %d" % mwg)
        params["MWG"] = mwg
        pyclblast.override_parameters(ctx.devices[0], 'Xgemm', 32, params)
        pyclblast.gemm(queue, m, n, k, cla, clb, clc, a_ld=k, b_ld=n, c_ld=n)
        assert np.allclose(clc.get(), a.dot(b)), "uh-oh, xgemm isn't behaving correctly"




# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import sys

import database.clblast as clblast


def get_best_results(database):
    """Retrieves the results with the lowest execution times"""
    sections_best = []
    for section in database["sections"]:
        section_best = {}

        # Stores all the section's meta data
        for attribute in section.keys():
            if attribute != "results":
                section_best[attribute] = section[attribute]
        if section_best["clblast_device_architecture"] == "" and section_best["clblast_device_vendor"] in clblast.VENDORS_WITH_ARCHITECTURE:
            section_best["clblast_device_architecture"] = clblast.DEVICE_ARCHITECTURE_DEFAULT

        # Find the best result
        parameters_best = None
        time_best = sys.float_info.max
        for result in section["results"]:
            if result["time"] < time_best:
                time_best = result["time"]
                parameters_best = result["parameters"]

        # Stores the best result
        section_best["results"] = [{"time": time_best, "parameters": parameters_best}]
        sections_best.append(section_best)

    return {"sections": sections_best}


def get_relative_bests(name, common_results, common_parameters, verbose=False):
    """Retrieves the parameters with the relative best execution time over different devices"""

    # Helper function
    def argmin(iterable):
        return min(enumerate(iterable), key=lambda x: x[1])[0]

    # Computes the sum of the execution times over the different devices
    performance_sums = []
    for parameters in common_parameters:
        performance_sum = sum([r["relative_time"] for r in common_results if r["parameters"] == parameters])
        performance_sums.append(performance_sum)

    # Retrieves the entry with the lowest time
    best_index = argmin(performance_sums)
    best_performance = performance_sums[best_index]
    best_parameters = common_parameters[best_index]

    # Completed, report and return the results
    if verbose:
        print("[database] " + str(name) + " with performance " + str(best_performance))
    return best_parameters



#!/usr/bin/env python

# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0.
# This file follows the PEP8 Python style guide and uses a max-width of 100 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import numpy as np
import pyopencl as cl
from pyopencl.array import Array
import pyclblast

# Settings for this sample:
batch_count = 2
dtype = 'float32'
alphas = [1.5, 1.0]
n = 4

print("# Setting up OpenCL")
ctx = cl.create_some_context()
queue = cl.CommandQueue(ctx)

print("# Setting up Numpy arrays")
x = np.random.rand(n * batch_count).astype(dtype=dtype)
y = np.random.rand(n * batch_count).astype(dtype=dtype)

print("# Batch offsets: next after each other")
x_offsets = [0, n]
y_offsets = [0, n]

print("# Setting up OpenCL arrays")
clx = Array(queue, x.shape, x.dtype)
cly = Array(queue, y.shape, y.dtype)
clx.set(x)
cly.set(y)

print("# Example level-1 batched operation: AXPY-batched")
assert len(alphas) == len(x_offsets) == len(y_offsets) == batch_count
pyclblast.axpyBatched(queue, n, clx, cly, alphas, x_offsets, y_offsets)
queue.finish()

print("# Full result for vector y: %s" % str(cly.get()))
for i in range(batch_count):
	result = alphas[i] * x[x_offsets[i]:x_offsets[i] + n] + y[y_offsets[i]:y_offsets[i] + n]
	print("# Expected result batch #%d: %s" % (i, str(result)))



#!/usr/bin/env python

# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import argparse
import os
import sys

from benchmark import benchmark_single, COMPARISONS


BENCHMARKS = ["axpy", "gemv", "gemm", "summary", "axpybatched", "gemmbatched", "gemmstridedbatched"]


def parse_arguments(argv):
    parser = argparse.ArgumentParser(description="Runs all (main) benchmarks in one go for a given device")
    parser.add_argument("-c", "--comparisons", default=[], nargs='+', help="The library(s) to compare against (choose from %s)" % COMPARISONS)
    parser.add_argument("-p", "--platform", required=True, type=int, help="The ID of the OpenCL platform to test on")
    parser.add_argument("-d", "--device", required=True, type=int, help="The ID of the OpenCL device to test on")
    parser.add_argument("-x", "--precision", type=int, default=32, help="The precision to test for (choose from 16, 32, 64, 3232, 6464")
    parser.add_argument("-l", "--load_from_disk", action="store_true", help="Increase verbosity of the script")
    parser.add_argument("-t", "--plot_title", default="", help="The title for the plots, defaults to benchmark name")
    parser.add_argument("-o", "--output_folder", default=os.getcwd(), help="Sets the folder for output plots (defaults to current folder)")
    parser.add_argument("-v", "--verbose", action="store_true", help="Increase verbosity of the script")
    cl_args = parser.parse_args(argv)
    return vars(cl_args)


def benchmark_all(comparisons, platform, device, precision, load_from_disk,
                  plot_title, output_folder, verbose):
    for bench in BENCHMARKS:
        from_disk = load_from_disk
        for tight_plot in [True, False]:  # two plots for a single benchmark
            benchmark_single(bench, comparisons, platform, device, None, precision, from_disk,
                             plot_title, tight_plot, output_folder, verbose)
            from_disk = True  # for the next plot of the same data


if __name__ == '__main__':
    parsed_arguments = parse_arguments(sys.argv[1:])
    benchmark_all(**parsed_arguments)




# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import os

# Type settings (also change in database_structure.hpp)
STRING_LENGTH = 50
PARAMETERS_LENGTH = 16

# Constants from the C++ code
VENDOR_DEFAULT = "default"
DEVICE_TYPE_DEFAULT = "All"
DEVICE_NAME_DEFAULT = "default"
DEVICE_NAME_DEFAULT_CONSTANT = "kDeviceNameDefault                                        "
DEVICE_ARCHITECTURE_DEFAULT = "default"

# List of attributes
DEVICE_TYPE_ATTRIBUTES = ["clblast_device_vendor", "clblast_device_type"]
DEVICE_ATTRIBUTES = ["clblast_device_name", "clblast_device_architecture",
                     "device_core_clock", "device_compute_units"]
KERNEL_ATTRIBUTES = ["precision", "kernel_family"]
ARGUMENT_ATTRIBUTES = ["arg_m", "arg_n", "arg_k", "arg_alpha", "arg_beta",
                       "arg_from", "arg_to", "arg_step",
                       "arg_channels", "arg_height", "arg_width", "arg_kernel_h", "arg_kernel_w",
                       "arg_num_kernels", "arg_batch_count"]
ATTRIBUTES = DEVICE_ATTRIBUTES + DEVICE_TYPE_ATTRIBUTES + KERNEL_ATTRIBUTES + ARGUMENT_ATTRIBUTES
GROUP_ATTRIBUTES = DEVICE_TYPE_ATTRIBUTES + KERNEL_ATTRIBUTES + ["kernel"] + ARGUMENT_ATTRIBUTES

# Other constants
VENDORS_WITH_ARCHITECTURE = ["AMD", "NVIDIA"]

def precision_to_string(precision):
    """Translates a precision number (represented as Python string) into a descriptive string"""
    if precision == "16":
        return "Half"
    elif precision == "32":
        return "Single"
    elif precision == "64":
        return "Double"
    elif precision == "3232":
        return "ComplexSingle"
    elif precision == "6464":
        return "ComplexDouble"
    else:
        raise("Unknown precision: " + precision)


def get_cpp_separator():
    """Retrieves a C++ comment separator"""
    return "// ================================================================================================="


def get_cpp_header(family, precision):
    """Retrieves the C++ header"""
    return ("\n" + get_cpp_separator() + """
// This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. It
// is auto-generated by the 'scripts/database/database.py' Python script.
//
// This file populates the database with best-found tuning parameters for the '%s%s' kernels.
//\n"""
            % (family.title(), precision)) + get_cpp_separator() + "\n"


def get_cpp_header_namespace():
    return "\nnamespace clblast {\n" + "namespace database {\n"


def get_cpp_footer():
    """Retrieves the C++ footer"""
    return "\n} // namespace database\n" + "} // namespace clblast\n"


def get_cpp_precision(family, precision):
    """Retrieves the C++ code for the start of a new precision"""
    precision_string = precision_to_string(precision)
    camelcase_name = family.title().replace("_", "")
    return("\nconst DatabaseEntry %s%s = {\n  \"%s\", Precision::k%s"
           % (camelcase_name, precision_string, camelcase_name, precision_string))


def get_cpp_device_vendor(vendor, device_type):
    """Retrieves the C++ code for the (default) vendor and device type"""
    if vendor == VENDOR_DEFAULT and device_type == DEVICE_TYPE_DEFAULT:
        return "    { // Default\n      kDeviceType%s, \"%s\", {\n" % (device_type, vendor)
    device_type_caps = device_type[0].upper() + device_type[1:]
    return "    { // %s %ss\n      kDeviceType%s, \"%s\", {\n" % (vendor, device_type, device_type_caps, vendor)


def get_cpp_family_includes(family, precisions):
    result = "\n"
    result += "#include \"database/kernels/%s/%s.hpp\"\n\n" % (family, family)
    for precision in precisions:
        result += "#include \"database/kernels/%s/%s_%s.hpp\"\n" % (family, family, precision)
    return result


def get_hpp_family_includes(family, precisions):
    result = "\n"
    result += "#include \"database/database_structure.hpp\"\n"
    result += "\n"
    result += "namespace clblast {\n"
    result += "namespace database {\n"
    result += "\n"
    camelcase_name = family.title().replace("_", "")
    for precision in precisions:
        precision_string = precision_to_string(precision)
        result += "extern const DatabaseEntry %s%s;\n" % (camelcase_name, precision_string)
    result += "\n"
    result += "} // namespace database\n"
    result += "} // namespace clblast\n"
    return result


def print_as_name(name):
    return "Name{\"%-50s\"}" % name.strip()[:STRING_LENGTH]


def get_kernel_database_results(kernel_database):
    """Retrieves the best result from a group of results. Asserts for valid data"""
    assert len(kernel_database) >= 1

    all_results = [item["results"] for item in kernel_database]

    best_results = all_results[0]
    for results in all_results:

        # Debugging in case of unexpected results
        length_assumption = (len(results) == 1)
        params_assumption = (sorted(results[0]["parameters"]) == sorted(best_results[0]["parameters"]))
        if not length_assumption or not params_assumption:
            print("[database] ERROR: Found %d kernel databases, expected 1" % len(kernel_database))
            all_keys = sorted([key for item in kernel_database for key in item.keys()])
            missing_keys = set([x for x in all_keys if all_keys.count(x) != len(kernel_database)])
            print("[database] All keys in databases: %s" % str(set(all_keys)))
            print("[database] Missing keys in one or more databases: %s" % str(missing_keys))
            for index, item in enumerate(kernel_database):
                print("[database] %d:" % index)
                print(item)
        assert length_assumption
        assert params_assumption

        if results[0]["time"] < best_results[0]["time"]:
            best_results = results

    return best_results


def print_cpp_database(database, output_dir):
    """Outputs the database as C++ code"""

    # Iterates over the kernel families
    kernel_families = sorted(set([s["kernel_family"] for s in database["sections"]]))
    for family_name in kernel_families:
        family_database = [s for s in database["sections"] if s["kernel_family"] == family_name]

        # Goes into a new path for each kernel family
        family_path = os.path.join(output_dir, family_name)

        # Loops over the different precision (e.g. 16, 32, 3232, 64, 6464)
        precisions = sorted(set([s["precision"] for s in database["sections"]]))  # Based on full database
        for precision in precisions:
            precision_database = [s for s in family_database if s["precision"] == precision]

            # Opens a new file for each precision
            full_path = os.path.join(family_path, family_name + "_" + precision + ".hpp")
            with open(full_path, 'w+') as f:
                f.write(get_cpp_header(family_name, precision))
                f.write('#include "database/kernels/' + family_name + '/' + family_name + '.hpp"')
                f.write(get_cpp_header_namespace())
                f.write(get_cpp_precision(family_name, precision))

                # In case there is nothing found at all (e.g. 16-bit): continue as if this was a
                #  precision of 32 but with the defaults only
                if len(precision_database) == 0:
                    print("[database] No results found for %s:%s, retrieving defaults from %s:32" %
                          (family_name, precision, family_name))
                    precision_database = [s for s in family_database if s["precision"] == "32"
                                          and s["clblast_device_vendor"] == VENDOR_DEFAULT
                                          and s["clblast_device_type"] == DEVICE_TYPE_DEFAULT
                                          and s["clblast_device_name"] == DEVICE_NAME_DEFAULT]

                # Discovers the parameters for this kernel
                parameter_names = []
                for example_data in precision_database:
                    for example_result in example_data["results"]:
                        parameter_names.extend([str(k) for k in example_result["parameters"].keys()])
                parameter_names = sorted(list(set(parameter_names)))
                parameter_names_as_string = ", ".join(['"%s"' % p for p in parameter_names])
                f.write(", {" + parameter_names_as_string + "}, {\n")

                # Loops over device vendors (e.g. AMD)
                device_vendors = sorted(set([s["clblast_device_vendor"] for s in precision_database]))
                for vendor in device_vendors:
                    vendor_database = [s for s in precision_database if s["clblast_device_vendor"] == vendor]

                    # Loops over device types (e.g. GPU)
                    device_types = sorted(set([s["clblast_device_type"] for s in vendor_database]))
                    for device_type in device_types:
                        type_database = [s for s in vendor_database if s["clblast_device_type"] == device_type]
                        f.write(get_cpp_device_vendor(vendor, device_type))

                        # Loops over every architecture of this vendor-type combination
                        architectures = sorted(set([s["clblast_device_architecture"] for s in type_database]))
                        if vendor in VENDORS_WITH_ARCHITECTURE:
                            architectures = [a for a in architectures if a != ""]
                        for architecture in architectures:
                            architecture_database = [s for s in type_database if s["clblast_device_architecture"] == architecture]
                            architecture_string = DEVICE_ARCHITECTURE_DEFAULT if architecture == "" else architecture
                            f.write("        { \"%s\", {\n" % architecture_string)

                            # Loops over every device of this vendor-type combination
                            devices = sorted(set([s["clblast_device_name"] for s in architecture_database]))
                            for device_name in devices:
                                device_database = [s for s in architecture_database if s["clblast_device_name"] == device_name]
                                device_name_as_string = print_as_name(device_name) if device_name != DEVICE_NAME_DEFAULT else DEVICE_NAME_DEFAULT_CONSTANT
                                device_name_cpp = "          { %s, Params{ " % device_name_as_string
                                f.write(device_name_cpp)

                                # Collects the parameters for this entry
                                parameters = []
                                parameter_index = 0
                                kernels = sorted(set([s["kernel"] for s in device_database]))
                                for kernel in kernels:
                                    kernel_database = [s for s in device_database if s["kernel"] == kernel]
                                    results = get_kernel_database_results(kernel_database)

                                    assert len(results) == 1
                                    new_parameters = results[0]["parameters"]
                                    for parameter_name in sorted(new_parameters):
                                        assert parameter_name == parameter_names[parameter_index]
                                        parameter_value = new_parameters[parameter_name]
                                        parameters.append(str(parameter_value))
                                        parameter_index += 1

                                # Appends zero's to complete the list
                                assert parameter_index <= PARAMETERS_LENGTH
                                for append_index in range(parameter_index, PARAMETERS_LENGTH):
                                    parameters.append("0")

                                # Prints the entry
                                f.write(", ".join(parameters))
                                f.write(" } },\n")

                            # Prints the architecture footer
                            f.write("        } },\n")

                        # Prints the vendor-type combination footer
                        f.write("      }\n    },\n")

                # Prints the precision footer
                f.write("  }\n};\n")

                # Prints the file footer
                f.write(get_cpp_footer())

            # Creates the combined family sources
            full_path = os.path.join(family_path, family_name + ".cpp")
            with open(full_path, 'w+') as f:
                f.write(get_cpp_header(family_name, ""))
                f.write(get_cpp_family_includes(family_name, precisions))

            # Creates the combined family includes header
            full_path = os.path.join(family_path, family_name + ".hpp")
            with open(full_path, 'w+') as f:
                f.write(get_cpp_header(family_name, ""))
                f.write(get_hpp_family_includes(family_name, precisions))



# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

import csv
import subprocess


def k(value):
    return value * 1024


def m(value):
    return value * 1024 * 1024


def float_to_kilo_mega(value):
    if value % 1024 or value <= 1024:
        return "%.0f" % value
    elif value % (1024 * 1024) or value <= (1024 * 1024):
        return "%.0fK" % (value / 1024.0)
    else:
        return "%.0fM" % (value / (1024.0 * 1024.0))


def powers_of_2(start, stop):
    while start <= stop:
        yield start
        start *= 2


def precision_to_letter(precision):
    if precision == 16:
        return "H"
    elif precision == 32:
        return "S"
    elif precision == 64:
        return "D"
    elif precision == 3232:
        return "C"
    elif precision == 6464:
        return "Z"
    else:
        return "X"


def run_binary(command, arguments):
    full_command = command + " " + " ".join(arguments)
    print("[benchmark] Calling binary: %s" % str(full_command))
    try:
        result = subprocess.Popen(full_command, shell=True, stdout=subprocess.PIPE).stdout.read()
        return result.decode("ascii")
    except OSError as e:
        print("[benchmark] Error while running the binary, got exception: %s" + str(e))
        return False


def parse_results(csv_data):
    csv_data = csv_data.split("\n")
    results = csv.DictReader(csv_data, delimiter=";", skipinitialspace=True)
    results = [r for r in results]
    for result in results:
        for key in result:
            if "i" in result[key]:
                continue
            else:
                result[key] = float(result[key]) if "." in result[key] else int(result[key])
    return results
